#include <cmath>
#include "AudioTransform.h"
#include "chuck_fft.h"
#include "WaterfallBuffer.h"
#include "Utility.h"

AudioTransform::AudioTransform(unsigned int buffer_size, unsigned int num_waterfalls)
{
    m_buffer_size = buffer_size;

    m_fft_size = (2 * buffer_size);
    m_current_sample = 0;

    m_magnitude_spectra = new WaterfallBuffer(buffer_size, num_waterfalls);

    graphics_buffer.resize(buffer_size);
    windowed_data.resize(buffer_size);

    fft_buffer.resize(m_fft_size);

    m_hanning_window.resize(buffer_size);
    hanning(&m_hanning_window[0], buffer_size);
    m_hamming_window.resize(buffer_size);
    hamming(&m_hamming_window[0], buffer_size);
    m_blackman_window.resize(buffer_size);
    blackman(&m_blackman_window[0], buffer_size);

    m_window_type = HANNING;

    m_new_data_is_available = false;

    pthread_mutex_init(&graphics_mutex, NULL);

    m_pitch_guess = "";
}

AudioTransform::~AudioTransform()
{
    pthread_mutex_unlock(&graphics_mutex);
    pthread_mutex_destroy(&graphics_mutex);

    delete m_magnitude_spectra;
}

void AudioTransform::start_consuming_data()
{
    pthread_mutex_lock(&graphics_mutex);
    m_new_data_is_available = false;
}

void AudioTransform::stop_consuming_data()
{
    pthread_mutex_unlock(&graphics_mutex);
}

bool AudioTransform::new_data_is_available()
{
    return m_new_data_is_available;
}

unsigned int AudioTransform::num_waterfalls()
{
    return m_magnitude_spectra->num_waterfalls();
}

std::vector<float>* AudioTransform::get_windowed_data()
{
    return &windowed_data;
}

std::vector<float>* AudioTransform::get_magnitude_spectrum(unsigned int index)
{
    return m_magnitude_spectra->get_buffer(index);
}

std::string AudioTransform::get_pitch_guess()
{
    return m_pitch_guess;
}

void AudioTransform::accept_samples(float* samples, unsigned int num_samples)
{
    unsigned int i;
    pthread_mutex_lock(&graphics_mutex);
    for (i = 0; i< num_samples; i++)
    {
        graphics_buffer[m_current_sample++] = samples[i];
        if (m_current_sample >= m_buffer_size)
        {
            m_current_sample = 0;
            transform_buffer();
        }
    }
    pthread_mutex_unlock(&graphics_mutex);
}

void AudioTransform::set_window_type(AUDIO_WINDOW_TYPE window_type)
{
    m_window_type = window_type;
}

void AudioTransform::transform_buffer()
{
    // Stats recommended in class as starting points:
    // 1024 for window size
    // Half the window size for the hop size
    // FFT size depends on the resolution you want, but consider 1024 or 2048
    // Window type - hamming
    unsigned int i;
    int samples;
    float frequency;
    complex* cbuf;
    std::vector<float>* magnitude_spectrum_buffer;

    for (i=0; i < m_buffer_size; i++)
    {
        switch (m_window_type)
        {
        case HAMMING:
            windowed_data[i] = graphics_buffer[i] * m_hamming_window[i];
            break;
        case HANNING:
            windowed_data[i] = graphics_buffer[i] * m_hanning_window[i];
            break;
        case BLACKMAN:
            windowed_data[i] = graphics_buffer[i] * m_blackman_window[i];
            break;
        case RECTANGULAR:
        default:
            windowed_data[i] = graphics_buffer[i];
            break;
        }
        fft_buffer[i] = windowed_data[i];
    }
    for (i=m_buffer_size; i<m_fft_size; i++)
    {
        fft_buffer[i] = 0;
    }

    // Take FFT and convert to magnitude spectrum
    rfft(&(fft_buffer[0]), m_fft_size/2, FFT_FORWARD );
    cbuf = (complex*)(&(fft_buffer[0]));
    magnitude_spectrum_buffer = m_magnitude_spectra->get_next_write_buffer();
    for(i = 0; i < m_buffer_size; i++ )
    {
        (*magnitude_spectrum_buffer)[i] = cmp_abs(cbuf[i]);
    }

    // Guess pitch based on unbiased autocorrelation: Make an
    // autocorrelation matrix, find the lag where it has the maximum
    // value, calculate based on this.
    float sum = 0.0;
    float max = 0.0;
    unsigned int n;
    unsigned int l;
    unsigned int min_lag = 10;
    unsigned int max_lag = (unsigned int)std::sqrt((float)m_buffer_size);
    samples = 0;
    for (l=min_lag; l < max_lag; l++)
    {
        sum=0;
        for (n=0; n < m_buffer_size - l; n++)
        {
            sum += graphics_buffer[n] * graphics_buffer[n+l];
        }
        sum /= (m_buffer_size - l);
        if (sum >= max)
        {
            samples = l;
            max = sum;
        }
    }
    // Threshold the outputs
    if (samples > 0 && max > 0.001)
    {
        frequency = (float)Utility::sample_rate / (float)samples;
        m_pitch_guess = Utility::midi_to_string(Utility::frequency_to_midi(frequency));
    }
    else
    {
        m_pitch_guess = "??";
    }

    m_new_data_is_available = true;
}
